// ****************************************************************
//
//   IBM Confidential
//   OCO Source Materials
//   <<PART NUMBER - 5724-B75>>
//   (C) Copyright IBM Corp. 2001 - All Rights Reserved.
//   US Government Users Restricted Rights - Use, duplication or disclosure
//   restricted by GSA ADP Schedule Contract with IBM Corp.
//
// ****************************************************************

package com.ibm.etools.marshall.util;

import java.io.*;

public class MarshallStringUtils
{
	private static final String copyright = "(c) Copyright IBM Corporation 2001.";

	public final static int STRING_JUSTIFICATION_LEFT = 0;
	public final static int STRING_JUSTIFICATION_RIGHT = 1;
	public final static int STRING_JUSTIFICATION_CENTER = 2;
	private static final char DBCS_SPACE_CHAR = '\u3000';
	private static final String DBCS_SPACE_STRING = "\u3000";

/**
 * TD has:
 * @param value java.lang.String
 * @param buffer byte[]
 * @param startingOffset int
 * @param size int
 * @param codePage java.lang.String
 * @param size int
 */

static public void marshallCharInfoBuffer (char value, byte[] buffer, int startingOffset, int size, String codePage)
	throws IllegalArgumentException
{
	String valueStr = (new Character(value)).toString();
	byte[] tb = null;

	try
	{
		tb = valueStr.getBytes (codePage);
	}
	catch (UnsupportedEncodingException exc)
	{
		throw new IllegalArgumentException(MarshallResource.instance().getString(MarshallResource.MARSHRT_STR_CODEPAGE_NOT_SUPPORTED, codePage));
	}

	if (tb.length > size)
		throw new IllegalArgumentException(MarshallResource.instance().getString(MarshallResource.MARSHRT_STR_LENGTH_NOT_VALID, valueStr, Integer.toString(size)));
	
	System.arraycopy (tb, 0, buffer, startingOffset, tb.length);

	for (int i = tb.length; i < size; i++)		// Pad rest of the bytes in buffer with zeros if there is space left
		buffer[startingOffset+i] = (byte)0x0;
	
	return;
}

/**
 * Java format of string may have single or double byte spaces (\u0020 or \u3000).
 *
 * Cobol PIC G  requires all characters (include the space character) to be double byte characters,
 * and there should be no shift characters, as Cobol does not use SI and SO bytes.
 *
 * Java format of a DBCS character, double byte space and then DBCS character for EBCDIC:  0e xx xx 40 40 xx xx 0f
 * Java format of a DBCS character, double byte space and then DBCS character for ASCII:   xx xx 81 40 xx xx
 *
 * Cobol format of a DBCS character, space and then DBCS character for EBCDIC:             xx xx 40 40 xx xx
 * Cobol format of a DBCS character, space and then DBCS character for ASCII:              xx xx 20 20 xx xx
 */

static public int marshallFixedLengthDBCSStringIntoBuffer (String value, byte[] buffer, int startingOffset, String codePage, int charSize, int size, int justification, String padCharStr, boolean DBCSOnly)
	throws IllegalArgumentException
{
	if (DBCSOnly)	// Convert any SBCS spaces to DBCS spaces before marshalling - can affect the size after data conversion.
		value = value.replace (' ', DBCS_SPACE_CHAR);

	// Convert the string to the DBCS codepage
	byte[] tb = null;

	try
	{
		tb = value.getBytes (codePage);
	}
	catch (UnsupportedEncodingException exc)
	{
		throw new IllegalArgumentException(MarshallResource.instance().getString(MarshallResource.MARSHRT_STR_CODEPAGE_NOT_SUPPORTED, codePage));
	}

	int bytesMarshalled = tb.length;
	int start = 0;

	if (DBCSOnly && (tb.length != 0))
	{
		if (tb[0] == 0x0e)	// Starts with shift in character - EBCDIC 	- Must remove SI and SO characters from the byte stream
		{					// (first and last character in the byte stream).
			if ((bytesMarshalled-2) > size)
				throw new IllegalArgumentException(MarshallResource.instance().getString(MarshallResource.MARSHRT_STR_LENGTH_NOT_VALID, value, Integer.toString(size)));

			start = 1;
			bytesMarshalled -= 2;
		}
		else		// Ascii codepage - need to massage the spaces in the byte stream
		{
			if (bytesMarshalled > size)
				throw new IllegalArgumentException(MarshallResource.instance().getString(MarshallResource.MARSHRT_STR_LENGTH_NOT_VALID, value, Integer.toString(size)));

			// Must change the byte pattern 0x81 0x40 (Java DBCS Space) to 0x20 0x20 (COBOL ascii DBCS Space)

			byte[] spaceBytes = null;
			byte fillByte = 0;
			try	{
				spaceBytes = DBCS_SPACE_STRING.getBytes (codePage);
				String tmp = " ";
				fillByte = tmp.getBytes(codePage)[0];		// Should be 0x20
			} catch (UnsupportedEncodingException exc) { }

			for (int i = 0; i < bytesMarshalled; i += 2)
			{
				if ((tb[i] == spaceBytes[0]) && (tb[i+1] == spaceBytes[1]))
				{
					tb[i] = fillByte;
					tb[i+1] = fillByte;
				}
			}
		}
	}
	else
	{
		// After conversion, DBCS characters could have made the buffer larger than
		// the maximum size of the String type.
		if (bytesMarshalled > size)
			throw new IllegalArgumentException(MarshallResource.instance().getString(MarshallResource.MARSHRT_STR_LENGTH_NOT_VALID, value, Integer.toString(size)));
	}

	int bytesToCopy = bytesMarshalled;

	// Copy the marshalled buffer into the byte array buffer
	switch (justification)
	{
		case STRING_JUSTIFICATION_LEFT:
			System.arraycopy (tb, start, buffer, startingOffset, bytesToCopy);
			break;

		case STRING_JUSTIFICATION_RIGHT :
			int offset = size - bytesToCopy;
			System.arraycopy (tb, start, buffer, startingOffset+offset, bytesToCopy);
			bytesMarshalled = size;
			break;

		case STRING_JUSTIFICATION_CENTER:
			offset = (size - bytesToCopy) / 2;
			System.arraycopy (tb, start, buffer, startingOffset+offset, bytesToCopy);
			bytesMarshalled = size;
			break;

		default:
			throw new IllegalArgumentException(MarshallResource.instance().getString(MarshallResource.MARSHRT_STR_JUSTIFICATION_NOT_VALID, Integer.toString(justification)));
	}

	// Pad out the string if nescessary
	if (padCharStr != null)
	{
		int padLength = size - bytesToCopy;

		if (padLength > 0)
		{
			bytesMarshalled = size;
			byte[] padArray = null;
			try
			{
				padArray = padCharStr.getBytes (codePage);
			}
			catch (UnsupportedEncodingException exc) {}

			switch (justification)
			{
				case STRING_JUSTIFICATION_LEFT:
					ConversionUtils.fillBytes (padArray[0], buffer, startingOffset+bytesToCopy, padLength);
					break;

				case STRING_JUSTIFICATION_RIGHT :
					ConversionUtils.fillBytes (padArray[0], buffer, startingOffset, padLength);
					break;

				case STRING_JUSTIFICATION_CENTER:
					// Need to pad the beginning and the end of the string
					int startBytes = padLength / 2;
					int endBytes = padLength - startBytes;
					ConversionUtils.fillBytes (padArray[0], buffer, startingOffset, startBytes);
					ConversionUtils.fillBytes (padArray[0], buffer, startingOffset+bytesToCopy, endBytes);
					break;
			}
		}
	}

	return (bytesMarshalled);
}

/**
 * TD has:
 * encoding -> codepage (String)
 * lengthEncoding -> lengthEncodingValue : Enumeration
 *   : fixedLength
 *   : lengthPrefixed
 *   : nullTerminated
 * maxLengthFormula -> max size formula as String
 *
 * Need to know if a padding character is present for marshalling.
 * @return The number of bytes marshalled.  If it is not left justified, or it is
 * padded out, it will be the maximum size of the field.
 * @param value java.lang.String
 * @param buffer byte[]
 * @param startingOffset int
 * @param codePage java.lang.String
 * @param size int
 */

static public int marshallFixedLengthStringIntoBuffer (String value, byte[] buffer, int startingOffset, String codePage, int size, int justification, String padCharStr)
	throws IllegalArgumentException
{
	byte[] tb = null;

	try
	{
		tb = value.getBytes (codePage);
	}
	catch (UnsupportedEncodingException exc)
	{
		throw new IllegalArgumentException(MarshallResource.instance().getString(MarshallResource.MARSHRT_STR_CODEPAGE_NOT_SUPPORTED, codePage));
	}

	int bytesMarshalled = tb.length;

	// After conversion DBCS characters could have made the buffer larger than
	// the maximum size of the String type.
	if (bytesMarshalled > size)
		throw new IllegalArgumentException(MarshallResource.instance().getString(MarshallResource.MARSHRT_STR_LENGTH_NOT_VALID, value, Integer.toString(size)));

	switch (justification)
	{
		case STRING_JUSTIFICATION_LEFT:
			System.arraycopy (tb, 0, buffer, startingOffset, tb.length);
			break;

		case STRING_JUSTIFICATION_RIGHT :
			int offset = size - tb.length;
			System.arraycopy (tb, 0, buffer, startingOffset+offset, tb.length);
			bytesMarshalled = size;
			break;

		case STRING_JUSTIFICATION_CENTER:
			bytesMarshalled = size;
			// Offset to copy at is startingOffset+(size-numBytes)/2
			offset = (size - tb.length) / 2;
			System.arraycopy (tb, 0, buffer, startingOffset+offset, tb.length);
			break;

		default:
			throw new IllegalArgumentException(MarshallResource.instance().getString(MarshallResource.MARSHRT_STR_JUSTIFICATION_NOT_VALID, Integer.toString(justification)));
	}

	if (padCharStr != null)		// Pad out the string
	{
		int padLength = size - tb.length;

		if (padLength > 0)
		{
			bytesMarshalled = size;
			byte[] padArray = null;
			try
			{
				padArray = padCharStr.getBytes (codePage);
			}
			catch (UnsupportedEncodingException exc) {}

			switch (justification)
			{
				case STRING_JUSTIFICATION_LEFT:
					ConversionUtils.fillBytes (padArray[0], buffer, startingOffset+tb.length, padLength);
					break;

				case STRING_JUSTIFICATION_RIGHT :
					ConversionUtils.fillBytes (padArray[0], buffer, startingOffset, padLength);
					break;

				case STRING_JUSTIFICATION_CENTER:
					// Need to pad the beginning and the end of the string
					int startBytes = padLength / 2;
					int endBytes = padLength - startBytes;
					ConversionUtils.fillBytes (padArray[0], buffer, startingOffset, startBytes);
					ConversionUtils.fillBytes (padArray[0], buffer, startingOffset+tb.length, endBytes);
					break;
			}
		}
	}

	return (bytesMarshalled);
}

/**
 * TD has:
 * encoding -> codepage (String)
 * lengthEncoding -> lengthEncodingValue : Enumeration
 *   : fixedLength
 *   : lengthPrefixed
 *   : nullTerminated
 * maxLengthFormula -> max size formula as String
 *
 * Need to know if a padding character is present for marshalling.
 * @param value java.lang.String
 * @param buffer byte[]
 * @param startingOffset int
 * @param codePage java.lang.String
 * @param size int
 */

static public void marshallLengthPrefixedDBCSStringIntoBuffer (String value, byte[] buffer, int startingOffset, String codePage, int charSize, int size, int justification, String padCharStr, boolean bigEndian, int prefixLength, boolean DBCSOnly)
	throws IllegalArgumentException
{
	int numBytes = marshallFixedLengthDBCSStringIntoBuffer (value, buffer, startingOffset+prefixLength,
							codePage, charSize, size-prefixLength, justification, padCharStr, DBCSOnly);

	// Put the integer length of the string into the buffer
	byte[] tb = null;

	switch (prefixLength)
	{
		case 4:
			tb = ConversionUtils.convertToByte (numBytes);
			break;

		case 2:
			tb = ConversionUtils.convertToByte ((short)numBytes);
			break;

		case 8:
			tb = ConversionUtils.convertToByte ((long)numBytes);
			break;

		case 1:
			tb = new byte[1];
			tb[0] = (byte)numBytes;
			break;		

		default:
			throw new IllegalArgumentException(MarshallResource.instance().getString(MarshallResource.MARSHRT_STR_PREFIX_LENGTH_NOT_VALID, Integer.toString(prefixLength)));
	}

	if (!bigEndian)
		ConversionUtils.endianSwap (tb);

	System.arraycopy (tb, 0, buffer, startingOffset, tb.length);

	return;
}

/**
 * TD has:
 * encoding -> codepage (String)
 * lengthEncoding -> lengthEncodingValue : Enumeration
 *   : fixedLength
 *   : lengthPrefixed
 *   : nullTerminated
 * maxLengthFormula -> max size formula as String
 *
 * Need to know if a padding character is present for marshalling.
 * @param value java.lang.String
 * @param buffer byte[]
 * @param startingOffset int
 * @param codePage java.lang.String
 * @param size int
 */

static public void marshallLengthPrefixedStringIntoBuffer (String value, byte[] buffer, int startingOffset,
			String codePage, int size, int justification, String padCharStr, boolean bigEndian, int prefixLength)
	throws IllegalArgumentException
{
	int numBytes = marshallFixedLengthStringIntoBuffer (value, buffer, startingOffset+prefixLength,
							codePage, size-prefixLength, justification, padCharStr);

	// Put the integer length of the string into the buffer
	byte[] tb = null;

	switch (prefixLength)
	{
		case 4:
			tb = ConversionUtils.convertToByte (numBytes);
			break;

		case 2:
			tb = ConversionUtils.convertToByte ((short)numBytes);
			break;

		case 8:
			tb = ConversionUtils.convertToByte ((long)numBytes);
			break;

		case 1:
			tb = new byte[1];
			tb[0] = (byte)numBytes;
			break;		

		default:
			throw new IllegalArgumentException(MarshallResource.instance().getString(MarshallResource.MARSHRT_STR_PREFIX_LENGTH_NOT_VALID, Integer.toString(prefixLength)));
	}

	if (!bigEndian)
		ConversionUtils.endianSwap (tb);

	System.arraycopy (tb, 0, buffer, startingOffset, tb.length);

	return;
}

/**
 * TD has:
 * encoding -> codepage (String)
 * lengthEncoding -> lengthEncodingValue : Enumeration
 *   : fixedLength
 *   : lengthPrefixed
 *   : nullTerminated
 * maxLengthFormula -> max size formula as String
 *
 * Need to know if a padding character is present for marshalling.
 * @param value java.lang.String
 * @param buffer byte[]
 * @param startingOffset int
 * @param codePage java.lang.String
 * @param size int
 */

static public void marshallNullTerminatedDBCSStringIntoBuffer (String value, byte[] buffer, int startingOffset, String codePage, int charSize, int size, boolean DBCSOnly)
	throws IllegalArgumentException
{
	if (!DBCSOnly)
	{
		marshallNullTerminatedDBCSStringCharByChar (value, buffer, startingOffset, codePage, charSize, size);
		return;
	}

	int numBytes = marshallFixedLengthDBCSStringIntoBuffer (value, buffer, startingOffset,
							codePage, charSize, size, STRING_JUSTIFICATION_LEFT, null, DBCSOnly);
	// Add the null character to the end of the string up to the size of the charSize.
	for (int i = 0; i < charSize; i++)
	{
		if (numBytes+i < size)
			buffer[startingOffset+numBytes+i] = 0;
		else
			break;
	}

	return;
}

// To handle C language wide characters.  If a single byte codepage is specified, then
// each wide character is padded out with nulls, so cannot do a straight codepage conversion.

static private void marshallNullTerminatedDBCSStringCharByChar (String value, byte[] buffer, int startingOffset, String codePage, int charSize, int size)
	throws IllegalArgumentException
{
	// Convert the string character by character
	char[] characters = value.toCharArray();

	// After conversion, DBCS characters could have made the buffer larger than
	// the maximum size of the String type.
	int maxChars = size / charSize;

	if (characters.length > maxChars)
		throw new IllegalArgumentException(MarshallResource.instance().getString(MarshallResource.MARSHRT_STR_LENGTH_NOT_VALID, value, Integer.toString(size)));

	// Convert the string to the DBCS codepage
	// Allocating the entire string size up front and copying converted characters into
	// the appropriate locations, automatically pads each character, and the end of the
	// string with nulls.
	byte[] tb = new byte[size];

	for (int i = 0; i < characters.length; i++)
	{
		Character currChar = new Character(characters[i]);
		try
		{
			byte[] currBytes = currChar.toString().getBytes (codePage);
			int currOffset = i * charSize;
			System.arraycopy (currBytes, 0, tb, currOffset, currBytes.length);
		}
		catch (UnsupportedEncodingException exc)
		{
			throw new IllegalArgumentException(MarshallResource.instance().getString(MarshallResource.MARSHRT_STR_CODEPAGE_NOT_SUPPORTED, codePage));
		}
	}

	// Copy the marshalled buffer into the byte array buffer
	System.arraycopy (tb, 0, buffer, startingOffset, tb.length);
}


/**
 * TD has:
 * encoding -> codepage (String)
 * lengthEncoding -> lengthEncodingValue : Enumeration
 *   : fixedLength
 *   : lengthPrefixed
 *   : nullTerminated
 * maxLengthFormula -> max size formula as String
 *
 * Need to know if a padding character is present for marshalling.
 * @param value java.lang.String
 * @param buffer byte[]
 * @param startingOffset int
 * @param codePage java.lang.String
 * @param size int
 */

static public void marshallNullTerminatedStringIntoBuffer (String value, byte[] buffer, int startingOffset, String codePage, int size)
	throws IllegalArgumentException
{
	byte[] tb = null;

	try
	{
		tb = value.getBytes (codePage);
	}
	catch (UnsupportedEncodingException exc)
	{
		throw new IllegalArgumentException(MarshallResource.instance().getString(MarshallResource.MARSHRT_STR_CODEPAGE_NOT_SUPPORTED, codePage));
	}

	// After conversion DBCS characters could have made the buffer larger than
	// the maximum size of the String type.
	if (tb.length >= size)
		throw new IllegalArgumentException(MarshallResource.instance().getString(MarshallResource.MARSHRT_STR_LENGTH_NOT_VALID, value, Integer.toString(size)));

	System.arraycopy (tb, 0, buffer, startingOffset, tb.length);

	// Add the null character to the end of the string.
	buffer[startingOffset+tb.length] = 0;

	return;
}

/**
 * TD has:
 * @param value java.lang.String
 * @param buffer byte[]
 * @param startingOffset int
 * @param size int
 * @param codePage java.lang.String
 * @param size int
 */

static public char unmarshallCharFromBuffer (byte[] buffer, int startingOffset, int size, String codePage)
	throws IllegalArgumentException
{
	String bufString = null;

	try
	{
		bufString = new String(buffer, startingOffset, size, codePage);
	}
	catch (UnsupportedEncodingException exc)
	{
		throw new IllegalArgumentException(MarshallResource.instance().getString(MarshallResource.MARSHRT_STR_CODEPAGE_NOT_SUPPORTED, codePage));
	}
	
	char[] charArray = bufString.toCharArray();

	return (charArray[0]);
}

/**
 * Java format may have single or double byte spaces.  (\u0020 or \u3000).
 * We will always convert to a double byte space.
 *
 * Cobol PIC G  requires spaces to be double byte.
 *
 * Java format of a DBCS character, double byte space and then DBCS character for EBCDIC:  0e xx xx 40 40 xx xx 0f
 * Java format of a DBCS character, double byte space and then DBCS character for ASCII:   xx xx 81 40 xx xx
 *
 * Cobol format of a DBCS character, space and then DBCS character for EBCDIC:             xx xx 40 40  xx xx
 * Cobol format of a DBCS character, space and then DBCS character for ASCII:              xx xx 20 20  xx xx
 *
 * When converting from COBOL to Java, we will always generate a double byte space.
 */

private static byte[] fixInboundDBCSBuffer (byte[] buffer, int startingOffset, String codePage, int size)
	throws IllegalArgumentException
{
	byte[] tb = null;
	byte[] spaceBytes = null;
	byte spaceChar = 0;
	try
	{
		spaceBytes = DBCS_SPACE_STRING.getBytes (codePage);
		String tmp = " ";
		spaceChar = tmp.getBytes(codePage)[0];		// Should be 0x20 for ascii 0x40 for EBCDIC
	}
	catch (UnsupportedEncodingException exc)
	{
		throw new IllegalArgumentException(MarshallResource.instance().getString(MarshallResource.MARSHRT_STR_CODEPAGE_NOT_SUPPORTED, codePage));
	}

	if (spaceChar == 0x40)	// EBCDIC - must add SI and SO characters to byte stream before conversion
	{
		tb = new byte[size+2];
		System.arraycopy (buffer, startingOffset, tb, 1, size);
		// Add SI and SO characters
		tb[0] = 0x0E;
		tb[tb.length-1] = 0x0F;
	}
	else	// Ascii - must massage DBCS space characters to Java format.
	{		// Must convert byte pattern 0x20 0x20 (COBOL) to 0x81 0x40 (Java)
		tb = new byte[size];
		System.arraycopy (buffer, startingOffset, tb, 0, size);
		for (int i = 0; i < tb.length; i += 2)
		{
			if ((tb[i] == spaceChar) && (tb[i+1] == spaceChar))
			{
				tb[i] = spaceBytes[0];
				tb[i+1] = spaceBytes[1];
			}
		}
	}

	return (tb);
}


static public String unmarshallFixedLengthDBCSStringFromBuffer (byte[] buffer, int startingOffset, String codePage, int size, boolean DBCSOnly)
	throws IllegalArgumentException
{
	byte[] tb = null;

	if (DBCSOnly)	// Fix inbound byte stream
	{
		tb = fixInboundDBCSBuffer (buffer, startingOffset, codePage, size);
		size = tb.length;
		startingOffset = 0;
	}
	else
		tb = buffer;

	String result = null;

	try
	{
		result = new String(tb, startingOffset, size, codePage);
	}
	catch (UnsupportedEncodingException exc)
	{
		throw new IllegalArgumentException(MarshallResource.instance().getString(MarshallResource.MARSHRT_STR_CODEPAGE_NOT_SUPPORTED, codePage));
	}
	
	return (result);
}

/**
 * TD has:
 * encoding -> codepage (String)
 * lengthEncoding -> lengthEncodingValue : Enumeration
 *   : fixedLength
 *   : lengthPrefixed
 *   : nullTerminated
 * maxLengthFormula -> max size formula as String
 *
 * Need to know if a padding character is present for marshalling.
 * @param value java.lang.String
 * @param buffer byte[]
 * @param startingOffset int
 * @param codePage java.lang.String
 * @param size int
 */

static public String unmarshallFixedLengthStringFromBuffer (byte[] buffer, int startingOffset, String codePage, int size)
	throws IllegalArgumentException
{
	// !!!!  ***** TEMP FIX *****
	//	REMOVE FOR GM

	if (buffer[startingOffset] == 0)
		return ("");

	// !!!!  ***** END TEMP FIX *****
	//	REMOVE FOR GM

	String result = null;

	try
	{
		result = new String(buffer, startingOffset, size, codePage);
	}
	catch (UnsupportedEncodingException exc)
	{
		throw new IllegalArgumentException(MarshallResource.instance().getString(MarshallResource.MARSHRT_STR_CODEPAGE_NOT_SUPPORTED, codePage));
	}

	return (result);
}

/**
 * TD has:
 * encoding -> codepage (String)
 * lengthEncoding -> lengthEncodingValue : Enumeration
 *   : fixedLength
 *   : lengthPrefixed
 *   : nullTerminated
 * maxLengthFormula -> max size formula as String
 *
 * Need to know if a padding character is present for marshalling.
 * @param value java.lang.String
 * @param buffer byte[]
 * @param startingOffset int
 * @param codePage java.lang.String
 * @param size int
 */

static public String unmarshallLengthPrefixedDBCSStringFromBuffer (byte[] buffer, int startingOffset, String codePage, int size, boolean bigEndian, int prefixLength, boolean DBCSOnly)
	throws IllegalArgumentException
{
	// Get length from buffer (first prefixLength bytes)

	byte intBytes[] = new byte[prefixLength];
	System.arraycopy (buffer, startingOffset, intBytes, 0, prefixLength);

	if (!bigEndian)
		ConversionUtils.endianSwap (intBytes);

	int length = 0;

	switch (prefixLength)
	{
		case 4:
			length = ConversionUtils.convertFromByteToInt (intBytes);
			break;

		case 2:
			length = (int)ConversionUtils.convertFromByteToShort (intBytes);
			break;

		case 8:
			length = (int)ConversionUtils.convertFromByteToLong (intBytes);
			break;
			
		case 1:
			length = (int)intBytes[0];
			break;

		default:
			throw new IllegalArgumentException(MarshallResource.instance().getString(MarshallResource.MARSHRT_STR_PREFIX_LENGTH_NOT_VALID, Integer.toString(prefixLength)));
	}

	if (length > (size - prefixLength))
		throw new IllegalArgumentException(MarshallResource.instance().getString(MarshallResource.MARSHRT_STR_PREFIX_LENGTH_NOT_VALID, Integer.toString(length)));

	byte[] tb = null;
	int offset = 0;

	if (DBCSOnly)	// Fix inbound byte stream
	{
		tb = fixInboundDBCSBuffer (buffer, startingOffset+prefixLength, codePage, length);
		length = tb.length;
	}
	else
	{
		tb = buffer;
		offset = startingOffset+prefixLength;
	}

	String result = null;

	try
	{
		result = new String(tb, offset, length, codePage);
	}
	catch (UnsupportedEncodingException exc)
	{
		throw new IllegalArgumentException(MarshallResource.instance().getString(MarshallResource.MARSHRT_STR_CODEPAGE_NOT_SUPPORTED, codePage));
	}

	return (result);
}


/**
 * TD has:
 * encoding -> codepage (String)
 * lengthEncoding -> lengthEncodingValue : Enumeration
 *   : fixedLength
 *   : lengthPrefixed
 *   : nullTerminated
 * maxLengthFormula -> max size formula as String
 *
 * Need to know if a padding character is present for marshalling.
 * @param value java.lang.String
 * @param buffer byte[]
 * @param startingOffset int
 * @param codePage java.lang.String
 * @param size int
 */

static public String unmarshallLengthPrefixedStringFromBuffer (byte[] buffer, int startingOffset, String codePage, int size, boolean bigEndian, int prefixLength)
	throws IllegalArgumentException
{
	// Get length from buffer (first prefixLength bytes)

	byte intBytes[] = new byte[prefixLength];
	System.arraycopy (buffer, startingOffset, intBytes, 0, prefixLength);

	if (!bigEndian)
		ConversionUtils.endianSwap (intBytes);

	int length = 0;

	switch (prefixLength)
	{
		case 4:
			length = ConversionUtils.convertFromByteToInt (intBytes);
			break;

		case 2:
			length = (int)ConversionUtils.convertFromByteToShort (intBytes);
			break;

		case 8:
			length = (int)ConversionUtils.convertFromByteToLong (intBytes);
			break;
			
		case 1:
			length = (int)intBytes[0];
			break;

		default:
			throw new IllegalArgumentException(MarshallResource.instance().getString(MarshallResource.MARSHRT_STR_PREFIX_LENGTH_NOT_VALID, Integer.toString(prefixLength)));
	}

	if (length > (size - prefixLength))
		throw new IllegalArgumentException(MarshallResource.instance().getString(MarshallResource.MARSHRT_STR_PREFIX_LENGTH_NOT_VALID, Integer.toString(length)));

	String result = null;

	try
	{
		result = new String(buffer, startingOffset+prefixLength, length, codePage);
	}
	catch (UnsupportedEncodingException exc)
	{
		throw new IllegalArgumentException(MarshallResource.instance().getString(MarshallResource.MARSHRT_STR_CODEPAGE_NOT_SUPPORTED, codePage));
	}

	return (result);
}

/**
 * TD has:
 * encoding -> codepage (String)
 * lengthEncoding -> lengthEncodingValue : Enumeration
 *   : fixedLength
 *   : lengthPrefixed
 *   : nullTerminated
 * maxLengthFormula -> max size formula as String
 *
 * Need to know if a padding character is present for marshalling.
 * @param value java.lang.String
 * @param buffer byte[]
 * @param startingOffset int
 * @param codePage java.lang.String
 * @param size int
 */

static public String unmarshallNullTerminatedDBCSStringFromBuffer (byte[] buffer, int startingOffset, String codePage, int size, boolean DBCSOnly, int charSize)
	throws IllegalArgumentException
{
	byte[] tb = null;

	if (DBCSOnly)	// Fix inbound byte stream
	{
		tb = fixInboundDBCSBuffer (buffer, startingOffset, codePage, size);
		size = tb.length;
		startingOffset = 0;
	}
	else
		return (unmarshallNullTerminatedDBCSStringCharByChar (buffer, startingOffset, codePage, size, charSize));
//		tb = buffer;

	String result = null;

	try
	{
		result = new String(tb, startingOffset, size, codePage);
	}
	catch (UnsupportedEncodingException exc)
	{
		throw new IllegalArgumentException(MarshallResource.instance().getString(MarshallResource.MARSHRT_STR_CODEPAGE_NOT_SUPPORTED, codePage));
	}

	int index = result.indexOf (0);		// Find first null character

	if (index != -1)		// There is a null character
		result = result.substring (0, index);

	return (result);
}

static public String unmarshallNullTerminatedDBCSStringFromBuffer (byte[] buffer, int startingOffset, String codePage, int size, boolean DBCSOnly)
	throws IllegalArgumentException
{
	return (unmarshallNullTerminatedDBCSStringFromBuffer (buffer, startingOffset, codePage, size, DBCSOnly, 2));
}

// To handle C language wide characters.  If a single byte codepage is specified, then
// each wide character is padded out with nulls, so cannot do a straight codepage conversion.

static private String unmarshallNullTerminatedDBCSStringCharByChar (byte[] buffer, int startingOffset, String codePage, int size, int charSize)
	throws IllegalArgumentException
{
	// Must extract out each character from the byte array, convert it, and append it to
	// the resultant string.
	int maxChars = size / charSize;
	StringBuffer sb = new StringBuffer(maxChars);

	for (int i = 0; i < maxChars; i++)
	{
		try
		{
			int currOffset = startingOffset + (i * charSize);
			String currChar = new String(buffer, currOffset, charSize, codePage);
			int index = currChar.indexOf (0);		// Find first null character
			if (index != -1)		// There is a null character
				currChar = currChar.substring (0, index);
			sb.append (currChar);
		}
		catch (UnsupportedEncodingException exc)
		{
			throw new IllegalArgumentException(MarshallResource.instance().getString(MarshallResource.MARSHRT_STR_CODEPAGE_NOT_SUPPORTED, codePage));
		}		
	}

	String result = sb.toString();
	int index = result.indexOf (0);		// Find first null character

	if (index != -1)		// There is a null character
		result = result.substring (0, index);

	return (result);
}


/**
 * TD has:
 * encoding -> codepage (String)
 * lengthEncoding -> lengthEncodingValue : Enumeration
 *   : fixedLength
 *   : lengthPrefixed
 *   : nullTerminated
 * maxLengthFormula -> max size formula as String
 *
 * Need to know if a padding character is present for marshalling.
 * @param value java.lang.String
 * @param buffer byte[]
 * @param startingOffset int
 * @param codePage java.lang.String
 * @param size int
 */

static public String unmarshallNullTerminatedStringFromBuffer (byte[] buffer, int startingOffset, String codePage, int size)
	throws IllegalArgumentException
{
	String result = null;

	try
	{
		result = new String(buffer, startingOffset, size, codePage);
	}
	catch (UnsupportedEncodingException exc)
	{
		throw new IllegalArgumentException(MarshallResource.instance().getString(MarshallResource.MARSHRT_STR_CODEPAGE_NOT_SUPPORTED, codePage));
	}

	int index = result.indexOf (0);		// Find first null character

	if (index != -1)		// There is a null character
		result = result.substring (0, index);

	return (result);
}

}